#!C:\bld\signac-flow_1538815040027\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'signac-flow==0.6.3','console_scripts','flow'
__requires__ = 'signac-flow==0.6.3'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('signac-flow==0.6.3', 'console_scripts', 'flow')()
    )
